package storage

import (
	"testing"

	"github.com/stretchr/testify/assert"
	"github.com/stretchr/testify/require"

	"xbus.io/go-xbus/v3/api"
	"xbus.io/xbus/storage"
	"xbus.io/xbus/testutils"
)

// ProcessStorageFactory is any factory of ProcessStorage
type ProcessStorageFactory func(bt testing.TB) (storage.ProcessStorage, func())

// ProcessStorageTests is a test suite for ProcessStorage implementations
func ProcessStorageTests(t *testing.T, factory ProcessStorageFactory) {
	var uuids []api.UUID
	for i := 0; i != 9; i++ {
		value, err := api.UUIDV4()
		if err != nil {
			t.Fatal(err)
		}
		uuids = append(uuids, value)
	}

	t.Run("Process Life Cycle", func(t *testing.T) {
		ps, clean := factory(t)
		defer clean()

		p := storage.Process{
			TriggerEmitterID:  uuids[0],
			TriggerEnvelopeID: uuids[1],
		}

		id, err := ps.CreateProcess(p)
		if err != nil {
			t.Fatal(err)
		}

		t.Run("Get", func(t *testing.T) {
			p1, ok := ps.GetProcess(id)
			if !ok {
				t.Fatal("Could not get the created process")
			}
			assert.False(t, p1.ID.IsZero())
			assert.Equal(t, uuids[0], p1.TriggerEmitterID)
			assert.Equal(t, uuids[1], p1.TriggerEnvelopeID)
		})
		t.Run("SetPipeline", func(t *testing.T) {
			ps.SetPipeline(id, uuids[2], storage.NodeOutputRef{NodeID: "source", Output: "default"})
			p1, ok := ps.GetProcess(id)
			if !ok {
				t.Fatal("Could not get the created process")
			}
			assert.Equal(t, uuids[2], p1.PipelineID)
		})
		t.Run("PurgeNonTerminated", func(t *testing.T) {
			ids, err := ps.PurgeProcess([]api.UUID{id})
			if err != nil {
				t.Fatal(err)
			}
			if len(ids) != 0 {
				t.Fatalf("Should not purge any process, purged: %s", ids)
			}
		})
		t.Run("SetStatus", func(t *testing.T) {
			if err := ps.SetStatus(id, api.Process_ERROR, ""); err != nil {
				t.Fatal(err)
			}
			assert.Equal(t, api.Process_ERROR.String(), ps.GetState(id).Status.String())
		})
		t.Run("Purge", func(t *testing.T) {
			ids, err := ps.PurgeProcess([]api.UUID{id})
			if err != nil {
				t.Fatalf("Expected no error, got: %s", err)
			}
			assert.Equal(t, []api.UUID{id}, ids)
			assert.Equal(t, 0, len(ps.Query(api.ProcessFilter{ID: []api.UUID{id}})))
		})
	})

	t.Run("Save/Load GroupID & CreateFrom", func(t *testing.T) {
		ps, clean := factory(t)
		defer clean()

		var rootProcessID api.UUID
		t.Run("standalone process", func(t *testing.T) {
			p := storage.Process{
				TriggerEmitterID:  uuids[0],
				TriggerEnvelopeID: uuids[1],
			}

			id, err := ps.CreateProcess(p)
			if !assert.NoError(t, err) {
				t.FailNow()
			}
			rootProcessID = id

			p1, ok := ps.GetProcess(id)
			if !assert.True(t, ok, "Could not get the created process") {
				t.FailNow()
			}
			assert.False(t, p1.ID.IsZero())
			assert.True(t, p1.GroupID.IsZero())
			assert.True(t, p1.ReplayOf.IsZero())
		})
		var childProcessID api.UUID
		t.Run("child process", func(t *testing.T) {
			p := storage.Process{
				TriggerEmitterID:  uuids[0],
				TriggerEnvelopeID: uuids[1],
				GroupID:           rootProcessID,
				ReplayOf:          rootProcessID,
			}

			id, err := ps.CreateProcess(p)
			if !assert.NoError(t, err) {
				t.FailNow()
			}
			childProcessID = id

			p1, ok := ps.GetProcess(id)
			if !assert.True(t, ok, "Could not get the created process") {
				t.FailNow()
			}
			assert.False(t, p1.ID.IsZero())
			assert.Equal(t, rootProcessID, p1.GroupID)
			assert.Equal(t, rootProcessID, p1.ReplayOf)
		})
		t.Run("child-of-child process", func(t *testing.T) {
			p := storage.Process{
				TriggerEmitterID:  uuids[0],
				TriggerEnvelopeID: uuids[1],
				GroupID:           rootProcessID,
				ReplayOf:          childProcessID,
			}

			id, err := ps.CreateProcess(p)
			if !assert.NoError(t, err) {
				t.FailNow()
			}

			p1, ok := ps.GetProcess(id)
			if !assert.True(t, ok, "Could not get the created process") {
				t.FailNow()
			}
			assert.False(t, p1.ID.IsZero())
			assert.Equal(t, rootProcessID, p1.GroupID)
			assert.Equal(t, childProcessID, p1.ReplayOf)
		})
	})

	t.Run("ResultAck", func(t *testing.T) {
		ps, clean := factory(t)
		defer clean()

		p := storage.Process{
			TriggerEmitterID:  uuids[0],
			TriggerEnvelopeID: uuids[1],
		}

		id, err := ps.CreateProcess(p)
		if err != nil {
			t.Fatal(err)
		}

		ps.AckResult(id)

		logEntries := ps.GetUnprocessedProcessLogEntries()

		assert.Equal(t, 1, len(logEntries))
		assert.Equal(t, id, logEntries[0].ProcessID)
		assert.Equal(t, true, logEntries[0].ResultAck)
	})

	t.Run("ProcessState", func(t *testing.T) {
		ps, clean := factory(t)
		defer clean()

		p := storage.Process{
			TriggerEmitterID:  uuids[0],
			TriggerEnvelopeID: uuids[1],
		}

		id, err := ps.CreateProcess(p)
		if err != nil {
			t.Fatal(err)
		}

		t.Run("Save ResultAcked", func(t *testing.T) {
			func() {
				updater := ps.GetStateUpdater(id)
				defer testutils.DeferedAssertNoError(t, updater.Unlock)

				state, err := updater.Get()
				if err != nil {
					t.Fatal(err)
				}
				assert.False(t, state.ResultAcked)
				state.ResultAcked = true
				if err := updater.Set(state); err != nil {
					t.Fatal(err)
				}
			}()

			state := ps.GetState(id)
			assert.True(t, state.ResultAcked)
		})

	})

	t.Run("Query", func(t *testing.T) {
		ps, clean := factory(t)
		defer clean()

		p := storage.Process{
			TriggerEmitterID:  uuids[0],
			TriggerEnvelopeID: uuids[1],
		}

		id, err := ps.CreateProcess(p)
		if err != nil {
			t.Fatal(err)
		}

		t.Run("Query by ResultAcked", func(t *testing.T) {
			// Make sure we do have a state
			// XXX this should not be necessary but the current sql implementation
			// is a little weak
			func() {
				updater := ps.GetStateUpdater(id)
				defer testutils.DeferedAssertNoError(t, updater.Unlock)

				state, err := updater.Get()
				if err != nil {
					t.Fatal(err)
				}
				assert.False(t, state.ResultAcked)
				if err := updater.Set(state); err != nil {
					t.Fatal(err)
				}
			}()

			t.Run("No filter", func(t *testing.T) {
				list := ps.Query(api.ProcessFilter{
					ResultAcked: []bool{},
				})

				if len(list) != 1 {
					t.Fatal("no process found")
				}

				assert.Equal(t, id, list[0].ID)
			})

			t.Run("Filter false", func(t *testing.T) {
				list := ps.Query(api.ProcessFilter{
					ResultAcked: []bool{false},
				})

				if len(list) != 1 {
					t.Fatal("no process found")
				}

				assert.Equal(t, id, list[0].ID)
			})

			t.Run("Filter true", func(t *testing.T) {
				list := ps.Query(api.ProcessFilter{
					ResultAcked: []bool{true},
				})
				assert.Equal(t, 0, len(list))
			})

		})
	})

	t.Run("QuerySummary", func(t *testing.T) {
		ps, clean := factory(t)
		defer clean()

		id, err := ps.CreateProcess(storage.Process{
			TriggerEmitterID:  uuids[0],
			TriggerEnvelopeID: uuids[1],
			PipelineID:        uuids[2],
		})
		require.NoError(t, err)
		require.NoError(t, ps.SetStatus(id, api.Process_RUNNING, "go"))

		id, err = ps.CreateProcess(storage.Process{
			TriggerEmitterID:  uuids[0],
			TriggerEnvelopeID: uuids[1],
			PipelineID:        uuids[2],
		})
		require.NoError(t, err)
		require.NoError(t, ps.SetStatus(id, api.Process_RUNNING, "go"))

		id, err = ps.CreateProcess(storage.Process{
			TriggerEmitterID:  uuids[0],
			TriggerEnvelopeID: uuids[1],
			PipelineID:        uuids[2],
		})
		require.NoError(t, err)
		require.NoError(t, ps.SetStatus(id, api.Process_ERROR, "go"))

		id, err = ps.CreateProcess(storage.Process{
			TriggerEmitterID:  uuids[0],
			TriggerEnvelopeID: uuids[1],
			PipelineID:        uuids[2],
		})
		require.NoError(t, err)
		require.NoError(t, ps.SetStatus(id, api.Process_DONE, "go"))

		id, err = ps.CreateProcess(storage.Process{
			TriggerEmitterID:  uuids[3],
			TriggerEnvelopeID: uuids[4],
			PipelineID:        uuids[5],
		})
		require.NoError(t, err)
		require.NoError(t, ps.SetStatus(id, api.Process_RUNNING, "go"))

		list := ps.Query(api.ProcessFilter{})
		require.Equal(t, 5, len(list))

		summary, err := ps.QuerySummary(api.ProcessFilter{})
		require.NoError(t, err)
		require.Equal(t, 4, len(summary))
		assert.Subset(t, summary, storage.ProcessSummary{
			{PipelineID: uuids[2], EmitterID: uuids[0], Status: api.Process_RUNNING, ResultAcked: false, Count: 2},
			{PipelineID: uuids[2], EmitterID: uuids[0], Status: api.Process_ERROR, ResultAcked: false, Count: 1},
			{PipelineID: uuids[2], EmitterID: uuids[0], Status: api.Process_DONE, ResultAcked: false, Count: 1},
			{PipelineID: uuids[5], EmitterID: uuids[3], Status: api.Process_RUNNING, ResultAcked: false, Count: 1},
		})
	})
}
