// Copyright 2012-2015 Apcera Inc. All rights reserved.
// Copyright 2016 Orus.io for the modified parts
// this file is covered by the Apache 2.0 licence

// Package xbus provides a few things that all xbus actors (services & clients)
// may need
package xbus

import (
	"io"
	"io/ioutil"
	"os"

	"github.com/rs/zerolog"
)

// LoggerOptions is the struct you need to pass to the GetLogger function
// it can be just created and passed untouched to the GetLogger if you need
// some quick logging functionality.
// it is recommended to at least set the Debug and Trace flags according to
// your taste though
type LoggerOptions struct {
	AppName string
	Logtime bool
	Debug   bool
	Trace   bool

	OutputMiddleware func(io.Writer) io.Writer
}

// Logger is the interface of xbus loggers
type Logger interface {
	// Log a notice statement
	Noticef(format string, v ...interface{})

	// Log a fatal error
	Fatalf(format string, v ...interface{})

	// Log an error
	Errorf(format string, v ...interface{})

	// Log a debug statement
	Debugf(format string, v ...interface{})

	// Log a trace statement
	Tracef(format string, v ...interface{})
}

// ZeroLogger is a Logger that gives access to an underlying zerolog.Logger
type ZeroLogger interface {
	Logger
	Zero() *zerolog.Logger
}

func autoColors() bool {
	stat, err := os.Stdout.Stat()
	if err != nil || (stat.Mode()&os.ModeCharDevice) == 0 {
		return false
	}
	return true
}

// GetLogger is a convenience function that returns a logger configured
// according to your options and ready to be used in your server
func GetLogger(opts *LoggerOptions) ZeroLogger {
	var out io.Writer = os.Stderr

	if autoColors() {
		out = zerolog.ConsoleWriter{Out: out}
	}

	if opts.OutputMiddleware != nil {
		out = opts.OutputMiddleware(out)
	}

	context := zerolog.
		New(out).
		With().
		Str("app", opts.AppName)
	if opts.Logtime {
		context = context.
			Timestamp()
	}
	log := context.Logger()
	if opts.Trace {
		log = log.Level(zerolog.TraceLevel)
	} else if opts.Debug {
		log = log.Level(zerolog.DebugLevel)
	} else {
		log = log.Level(zerolog.InfoLevel)
	}

	return zeroLogger{&log}
}

// WrapZeroLogger wraps a zerolog.Logger
func WrapZeroLogger(log zerolog.Logger) ZeroLogger {
	return zeroLogger{&log}
}

// zeroLogger implements the ZeroLogger interface on top of a zerolog.Logger
type zeroLogger struct {
	*zerolog.Logger
}

func (l zeroLogger) Zero() *zerolog.Logger {
	return l.Logger
}

// Noticef logs a 'Info' message
func (l zeroLogger) Noticef(format string, v ...interface{}) {
	l.Info().Msgf(format, v...)
}

// Fatalf logs a fatal error
func (l zeroLogger) Fatalf(format string, v ...interface{}) {
	l.Fatal().Msgf(format, v...)
}

// Errorf logs an error
func (l zeroLogger) Errorf(format string, v ...interface{}) {
	l.Error().Msgf(format, v...)
}

// Debugf logs a debug statement
func (l zeroLogger) Debugf(format string, v ...interface{}) {
	l.Debug().Msgf(format, v...)
}

// Tracef logs a trace statement
func (l zeroLogger) Tracef(format string, v ...interface{}) {
	l.Trace().Msgf(format, v...)
}

// GetZeroLogger returns the zerolog.Logger behind a Logger
func GetZeroLogger(logger Logger) *zerolog.Logger {
	return logger.(ZeroLogger).Zero()
}

// NewPrefixedLogger PrefixedLogger constructor
func NewPrefixedLogger(prefix string, logger Logger) ZeroLogger {
	return LoggerWithStr(logger, "prefix", prefix)
}

// LoggerWithStr returns a new logger on witch the given value is added to the
// context
func LoggerWithStr(logger Logger, name, value string) ZeroLogger {
	return WrapZeroLogger(
		GetZeroLogger(logger).
			With().
			Str(name, value).
			Logger())
}

// NewNoopLogger returns a NoopLogger
func NewNoopLogger() ZeroLogger {
	log := zerolog.New(ioutil.Discard)
	return zeroLogger{&log}
}
